#!/usr/bin/env python

"""Tests for `safpis` package."""


from unittest import TestCase, mock
from click.testing import CliRunner

from safpis.api import SafpisAPI, APIKeyMissing
from safpis.models import Brand, Fuel, Site, SitePrice
from safpis import cli
from geopy.distance import Distance
from datetime import datetime
from decimal import Decimal
import os
import configparser


class TestSafpis(TestCase):
    """Tests for `safpis` package."""

    def setUp(self):
        secret_config_file = "secrets.cfg"
        if os.path.exists(secret_config_file):
            config = configparser.ConfigParser()
            config.read(secret_config_file)
            os.environ["SAFPIS_SUBSCRIBER_TOKEN"] = config["TEST"][
                "SAFPIS_SUBSCRIBER_TOKEN"
            ]

        self.site_dict = {
            "S": "61205460",
            "A": "11 Vader Street",
            "N": "OTR Dry Creek",
            "B": 169,
            "P": "5094",
            "G1": 170227225,
            "G2": 189,
            "G3": 4,
            "G4": 0,
            "G5": 0,
            "Lat": -34.819297,
            "Lng": 138.592116,
            "M": "2023-12-27T09:15:01.100",
            "GPI": "ChIJKy0p_ra3sGoRaWz3bT-5iEk",
            "MO": "00:00",
            "MC": "23:59",
            "TO": "00:00",
            "TC": "23:59",
            "WO": "00:00",
            "WC": "23:59",
            "THO": "00:00",
            "THC": "23:59",
            "FO": "00:00",
            "FC": "23:59",
            "SO": "00:00",
            "SC": "23:59",
            "SUO": "00:00",
            "SUC": "23:59",
        }
        self.site_prices_dict = {
            "SiteId": 61501045,
            "FuelId": 14,
            "CollectionMethod": "T",
            "TransactionDateUtc": "2021-01-06T22:55:00",
            "Price": 1356.0,
        }

    def tearDown(self):
        """Tear down test fixtures, if any."""

    @mock.patch.dict(os.environ, {"fake_key": "fake"}, clear=True)
    def test_token_None(self):
        with self.assertRaises(APIKeyMissing):
            SafpisAPI()

    @mock.patch.dict(os.environ, {"SAFPIS_SUBSCRIBER_TOKEN": ""})
    def test_token_empty(self):
        with self.assertRaises(APIKeyMissing):
            SafpisAPI()

    def test_token_valid(self):
        safpis_api = SafpisAPI()
        self.assertIsInstance(safpis_api, SafpisAPI)

    @mock.patch.dict(os.environ, {"SAFPIS_SUBSCRIBER_TOKEN": "INVALID-TOKEN"})
    def test_token_invalid(self):
        safpis_api = SafpisAPI()
        self.assertIsInstance(safpis_api, SafpisAPI)

    def test_get_country_brands(self):
        safpis_api = SafpisAPI()
        response = safpis_api.get_country_brands()
        self.assertIsInstance(response, dict)
        self.assertIn("Brands", response)

    @mock.patch.dict(os.environ, {"SAFPIS_SUBSCRIBER_TOKEN": "INVALID-TOKEN"})
    def test_get_country_brands_invalid_token(self):
        safpis_api = SafpisAPI()
        with self.assertRaises(Exception):
            safpis_api.get_country_brands()

    def test_get_country_geographic_regions(self):
        safpis_api = SafpisAPI()
        response = safpis_api.get_country_geographic_regions()
        self.assertIsInstance(response, dict)
        self.assertIn("GeographicRegions", response)

    def test_get_country_fuel_types(self):
        safpis_api = SafpisAPI()
        response = safpis_api.get_country_fuel_types()
        self.assertIsInstance(response, dict)
        self.assertIn("Fuels", response)

    def test_get_full_site_details(self):
        safpis_api = SafpisAPI()
        response = safpis_api.get_full_site_details()
        self.assertIsInstance(response, dict)
        self.assertIn("S", response)

    def test_get_sites_prices(self):
        safpis_api = SafpisAPI()
        response = safpis_api.get_sites_prices()
        self.assertIsInstance(response, dict)
        self.assertIn("SitePrices", response)

    def test_brands_by_id(self):
        safpis_api = SafpisAPI()
        brands = safpis_api.brands_by_id(id=2)
        self.assertIsInstance(brands, list)
        self.assertIsInstance(brands[0], Brand)
        self.assertEqual(brands[0].Name, "Caltex")

    def test_brands_by_id_non_existant(self):
        safpis_api = SafpisAPI()
        with self.assertRaises(Exception):
            safpis_api.brands_by_id(id=9999999)

    def test_brands_by_name(self):
        safpis_api = SafpisAPI()
        brands = safpis_api.brands_by_name(name="Caltex")
        self.assertIsInstance(brands, list)
        self.assertIsInstance(brands[0], Brand)
        self.assertEqual(brands[0].BrandId, 2)

    def test_brands_by_name_non_existant(self):
        safpis_api = SafpisAPI()
        with self.assertRaises(Exception):
            safpis_api.brands_by_name(name="Non-existent")

    def test_fuels_by_id(self):
        safpis_api = SafpisAPI()
        fuels = safpis_api.fuels_by_id(id=2)
        self.assertIsInstance(fuels, list)
        self.assertIsInstance(fuels[0], Fuel)
        self.assertEqual(fuels[0].Name, "Unleaded")

    def test_fuels_by_id_non_existant(self):
        safpis_api = SafpisAPI()
        with self.assertRaises(Exception):
            safpis_api.fuels_by_id(id=9999999)

    def test_fuels_by_name(self):
        safpis_api = SafpisAPI()
        fuels = safpis_api.fuels_by_name(name="e10")
        self.assertIsInstance(fuels, list)
        self.assertIsInstance(fuels[0], Fuel)
        self.assertEqual(fuels[0].FuelId, 12)

    def test_fuels_by_name_non_existant(self):
        safpis_api = SafpisAPI()
        with self.assertRaises(Exception):
            safpis_api.fuels_by_name(name="Non-existent")

    def test_site_instantiation_M_with_subsecond(self):
        site = Site(**self.site_dict)
        self.assertIsInstance(site, Site)

    def test_site_instantiation_M_without_subsecond(self):
        self.site_dict["M"] = "2023-12-27T09:15:01"
        site = Site(**self.site_dict)
        self.assertIsInstance(site, Site)

    def test_site_instantiation_MO_as_None(self):
        self.site_dict["MO"] = None
        site = Site(**self.site_dict)
        self.assertIsInstance(site, Site)

    def test_site_instantiation_MO_as_empty(self):
        self.site_dict["MO"] = ""
        site = Site(**self.site_dict)
        self.assertIsInstance(site, Site)

    def test_site_is_open(self):
        self.site_dict["MO"] = "06:00"
        self.site_dict["MC"] = "23:59"
        site = Site(**self.site_dict)
        self.assertTrue(site.is_open(datetime(2023, 12, 25, 6, 0, 0)))
        self.assertFalse(site.is_open(datetime(2023, 12, 25, 4, 0, 0)))
        self.assertTrue(site.is_open(datetime(2023, 12, 25, 23, 59, 0)))
        self.assertFalse(site.is_open(datetime(2023, 12, 25, 0, 0, 0)))
        self.assertTrue(site.is_open())

    def test_sites_by_id(self):
        safpis_api = SafpisAPI()
        sites = safpis_api.sites_by_id(id=61205460)
        self.assertIsInstance(sites, list)
        self.assertIsInstance(sites[0], Site)
        self.assertEqual(sites[0].N, "OTR Dry Creek")

    def test_sites_by_id_non_existant(self):
        safpis_api = SafpisAPI()
        with self.assertRaises(Exception):
            safpis_api.sites_by_id(id=9999999)

    def test_sites_by_name(self):
        safpis_api = SafpisAPI()
        sites = safpis_api.sites_by_name(name="OTR Dry Creek")
        self.assertIsInstance(sites, list)
        self.assertIsInstance(sites[0], Site)
        self.assertEqual(sites[0].S, 61205460)

    def test_sites_by_name_non_existant(self):
        safpis_api = SafpisAPI()
        with self.assertRaises(Exception):
            safpis_api.sites_by_name(name="Non-existent")

    def test_site_distance_between(self):
        site = Site(**self.site_dict)
        dist = site.distance_between(-34.819297, 138.592116)
        self.assertIsInstance(dist, Distance)
        self.assertEqual(dist.km, 0)

    def test_sites_by_distance(self):
        safpis_api = SafpisAPI()
        sites = safpis_api.sites_by_distance(
            latitude=-34.819297,
            longitude=138.592116,
            max_distance=Distance(kilometers=10),
        )
        self.assertIsInstance(sites, list)
        self.assertIsInstance(sites[0], Site)
        self.assertEqual(sites[0].N, "OTR Dry Creek")

    def test_sites_open(self):
        safpis_api = SafpisAPI()
        sites = safpis_api.sites_open(datetime(2023, 12, 25, 4, 0, 0))
        self.assertIsInstance(sites, list)
        self.assertIsInstance(sites[0], Site)

    def test_site_price(self):
        site_price = SitePrice(**self.site_prices_dict)
        self.assertIsInstance(site_price, SitePrice)
        self.assertIsInstance(site_price.Price.amount, Decimal)
        self.assertEqual(site_price.Price.amount, Decimal("1356"))

    def test_sites_by_cheapest_fuel_type(self):
        safpis_api = SafpisAPI()
        site_prices = safpis_api.sites_by_cheapest_fuel_type(fuel_type="e10")
        self.assertIsInstance(site_prices, list)
        self.assertIsInstance(site_prices[0], SitePrice)

    def test_command_line_interface(self):
        """Test the CLI."""
        runner = CliRunner()
        result = runner.invoke(cli.main)
        assert result.exit_code == 0
        assert "safpis.cli.main" in result.output
        help_result = runner.invoke(cli.main, ["--help"])
        assert help_result.exit_code == 0
        assert "--help  Show this message and exit." in help_result.output



import os
import requests_cache
import datetime
from safpis.models import Brand, Fuel, Site, SitePrice
from geopy.distance import Distance


class SafpisAPI:
    """This is a class for interacting with the SAFPIS REST API."""

    def __init__(self) -> None:
        """Constructor method"""
        self.base_url = (
            "https://" "fppdirectapi-prod.safuelpricinginformation.com.au"
        )
        self.__country_id = 21  # 21 = Australia
        self.__geo_region_level = 3  # 3 = States
        self.__geo_region_id = 4  # 4 = South Australia

        try:
            subscriber_token = os.environ.get("SAFPIS_SUBSCRIBER_TOKEN")
            if not subscriber_token:
                raise APIKeyMissing(
                    "Environmental variable SAFPIS_SUBSCRIBER_TOKEN is set "
                    "but empty."
                )
        except KeyError:
            raise APIKeyMissing(
                "Environmental variable SAFPIS_SUBSCRIBER_TOKEN is not set."
            )

        self.__token = subscriber_token

        self.cached_session_day = requests_cache.CachedSession(
            "safpis_cache_day",
            use_cache_dir=True,
            cache_control=True,
            expire_after=datetime.timedelta(days=1),
            allowable_codes=[200, 400],
            allowable_methods=["GET"],
            stale_if_error=False,
        )

        self.cached_session_minute = requests_cache.CachedSession(
            "safpis_cache_minute",
            use_cache_dir=True,
            cache_control=True,
            expire_after=datetime.timedelta(minutes=1),
            allowable_codes=[200, 400],
            allowable_methods=["GET"],
            stale_if_error=False,
        )

        self.headers = {
            "Content-Type": "application/json",
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate, br",
            "Authorization": f"FPDAPI SubscriberToken={self.__token}",
        }

    def __call_api(self, url: str, params: dict, cache: str = "day"):
        """Dispatches requests to the SAFPIS REST API, caching responses to
        avoid undue load on the service as required by the SAFPIS API (OUT)
        Guide.

        :param url: The path and endpoint for the request.
        :type url: str
        :param params: Request Parameters.
        :type params: dict
        :param cache: The cache used for the request, defaults to 'day'.  It
             can be 'day' or 'minute'.
        :type cache: str
        :raises ValueError: if :param cache: value is not recognised.
        :raises HTTPError: if the HTTP response is a 4XX client or 5XX server
            error response.
        :return: The request response.
        :rtype: requests.Response
        """
        valid_cache = {"day", "minute"}
        if cache not in valid_cache:
            raise ValueError("cache must be one of %r." % valid_cache)

        if cache == "day":
            response = self.cached_session_day.get(
                url,
                headers=self.headers,
                params=params,
            )
        elif cache == "minute":
            response = self.cached_session_minute.get(
                url,
                headers=self.headers,
                params=params,
            )

        response.raise_for_status()

        return response

    def get_country_brands(self, countryId: int = 21):
        """Sends a request to the GetCountryBrands endpoint, caching responses
        for a day.

        :param countryId: The ID of the country for which fuel brands are being
            requested, defaults to 21 (Australia).
        :type countryId: int
        :return: The json-encoded content of the response.
        """
        url = f"{self.base_url}/Subscriber/GetCountryBrands"
        params = {"countryId": countryId}

        response = self.__call_api(
            url,
            params=params,
        )

        return response.json()

    def get_country_geographic_regions(self, countryId: int = 21):
        """Sends a request to the GetCountryGeographicRegions endpoint, caching
        responses for a day.

        :param countryId: The ID of the country for which geographic regions
            are being requested, defaults to 21 (Australia).
        :type countryId: int
        :return: The json-encoded content of the response.
        """
        url = f"{self.base_url}/Subscriber/GetCountryGeographicRegions"
        params = {"countryId": countryId}

        response = self.__call_api(
            url,
            params=params,
        )

        return response.json()

    def get_country_fuel_types(self, countryId: int = 21):
        """Sends a request to the GetCountryFuelTypes endpoint, caching
        responses for a day.

        :param countryId: The ID of the country for which fuel types are being
            requested, defaults to 21 (Australia).
        :type countryId: int
        :return: The json-encoded content of the response.
        """
        url = f"{self.base_url}/Subscriber/GetCountryFuelTypes"
        params = {"countryId": countryId}

        response = self.__call_api(
            url,
            params=params,
        )

        return response.json()

    def get_full_site_details(
        self,
        countryId: int = 21,
        GeoRegionLevel: int = 3,
        GeoRegionId: int = 4,
    ):
        """Sends a request to the GetFullSiteDetails endpoint, caching
        responses for a day.

        :param countryId: The ID of the country for which site details are
            being requested, defaults to 21 (Australia).
        :type countryId: int
        :param GeoRegionLevel: The level of the geographic region for which
            site details are being requested, defaults to 3 (states).
        :type GeoRegionLevel: int
        :param GeoRegionId: The ID of the geographic region for which site
            details are being requested, defaults to 4 (South Australia).
        :type GeoRegionId: int
        :return: The json-encoded content of the response.
        """
        url = f"{self.base_url}/Subscriber/GetFullSiteDetails"
        params = {
            "countryId": countryId,
            "geoRegionLevel": GeoRegionLevel,
            "geoRegionId": GeoRegionId,
        }

        response = self.__call_api(
            url,
            params=params,
        )

        return response.json()

    def get_sites_prices(
        self,
        countryId: int = 21,
        GeoRegionLevel: int = 3,
        GeoRegionId: int = 4,
    ):
        """Sends a request to the GetSitesPrices endpoint, caching
        responses for a minute.

        :param countryId: The ID of the country for which site prices are
            being requested, defaults to 21 (Australia).
        :type countryId: int
        :param GeoRegionLevel: The level of the geographic region for which
            site prices are being requested, defaults to 3 (states).
        :type GeoRegionLevel: int
        :param GeoRegionId: The ID of the geographic region for which site
            prices are being requested, defaults to 4 (South Australia).
        :type GeoRegionId: int
        :return: The json-encoded content of the response.
        """
        url = f"{self.base_url}/Price/GetSitesPrices"
        params = {
            "countryId": countryId,
            "geoRegionLevel": GeoRegionLevel,
            "geoRegionId": GeoRegionId,
        }

        response = self.__call_api(
            url,
            params=params,
            cache="minute",
        )

        return response.json()

    def brands_by_id(self, id: int):
        brands = [
            Brand(**brand)
            for brand in self.get_country_brands()["Brands"]
            if brand["BrandId"] == id
        ]
        if not brands:
            raise ValueError(f"No brand found for id: {id}")
        return brands

    def brands_by_name(self, name: str):
        brands = [
            Brand(**brand)
            for brand in self.get_country_brands()["Brands"]
            if brand["Name"] == name
        ]
        if not brands:
            raise ValueError(f"No brand found for name: {name}")
        return brands

    def fuels_by_id(self, id: int):
        fuels = [
            Fuel(**fuel)
            for fuel in self.get_country_fuel_types()["Fuels"]
            if fuel["FuelId"] == id
        ]
        if not fuels:
            raise ValueError(f"No fuel type found for id: {id}")
        return fuels

    def fuels_by_name(self, name: str):
        fuels = [
            Fuel(**fuel)
            for fuel in self.get_country_fuel_types()["Fuels"]
            if fuel["Name"] == name
        ]
        if not fuels:
            raise ValueError(f"No fuel type found for name: {name}")
        return fuels

    def sites_by_id(self, id: int):
        sites = [
            Site(**site)
            for site in self.get_full_site_details()["S"]
            if site["S"] == id
        ]
        if not sites:
            raise ValueError(f"No sites found for id: {id}")
        return sites

    def sites_by_name(self, name: str):
        sites = [
            Site(**site)
            for site in self.get_full_site_details()["S"]
            if site["N"] == name
        ]
        if not sites:
            raise ValueError(f"No sites found for name: {name}")
        return sites

    def sites_by_distance(
        self, latitude: float, longitude: float, max_distance: Distance
    ):
        sites = [Site(**site) for site in self.get_full_site_details()["S"]]
        filtered_sites = filter(
            lambda site: site.distance_between(latitude, longitude)
            <= max_distance,
            sites,
        )
        sorted_sites = sorted(
            filtered_sites,
            key=lambda site: site.distance_between(latitude, longitude),
        )

        return sorted_sites

    def sites_open(self, datetime: datetime):
        sites = [Site(**site) for site in self.get_full_site_details()["S"]]
        filtered_sites = filter(lambda site: site.is_open(datetime), sites)

        return list(filtered_sites)

    def sites_by_cheapest_fuel_type(self, fuel_type: str):
        """
        Function to return a list of sites, ordered by price
        :param fuel_type: Fuel type. Type - str
        :return: List of Site objects
        """
        site_prices = [
            SitePrice(**site_price)
            for site_price in self.get_sites_prices()["SitePrices"]
        ]
        fuel_id = self.fuels_by_name(name=fuel_type)[0].FuelId
        filtered_site_prices = filter(
            lambda site_price: site_price.FuelId == fuel_id,
            site_prices,
        )
        sorted_site_prices = sorted(
            filtered_site_prices,
            key=lambda site_price: site_price.Price.amount,
        )

        return sorted_site_prices


class APIKeyMissing(Exception):
    """Exception for a missing SAFPIS Subscriber Token.

    The SAFPIS Subscriber Token should be specified in the environmental
    variable 'SAFPIS_SUBSCRIBER_TOKEN'.
    """

    pass



from datetime import datetime, time
import geopy.distance
from money import Money
from dataclasses import dataclass, field
from typing import Optional
from dateutil import parser
from decimal import Decimal


@dataclass
class Brand:
    """Brand returned ...
    :param BrandId: Unique Brand ID
    :param Name: Brand's name
    """

    BrandId: int
    Name: str


@dataclass
class Region:
    """Region returned ...
    :param GeoRegionId: Unique Region ID
    :param GeoRegionLevel: The Level (3 = State, 2 = City and 1 = Suburb)
    :param Name: Name of the Region
    :param Abbrev: Abbreviated form of the Region's Name
    :param GeoRegionParentId: ID of the Region's parent
    """

    GeoRegionId: int
    GeoRegionLevel: int
    Name: int
    Abbrev: str
    GeoRegionParentId: Optional[int] = field(default=None)


@dataclass
class Fuel:
    """Fuel returned ...
    :param FuelId: Unique Fuel Type ID
    :param Name: name of the Fuel Type
    """

    FuelId: int
    Name: str


@dataclass
class Site:
    S: int
    A: str
    N: str
    B: int
    P: int
    G1: int
    G2: int
    G3: int
    G4: int
    G5: int
    Lat: float
    Lng: float
    M: datetime
    GPI: str
    MO: time
    MC: time
    TO: time
    TC: time
    WO: time
    WC: time
    THO: time
    THC: time
    FO: time
    FC: time
    SO: time
    SC: time
    SUO: time
    SUC: time

    def __post_init__(self):
        if isinstance(self.M, str):
            try:
                self.M = datetime.strptime(self.M, "%Y-%m-%dT%H:%M:%S.%f")
            except ValueError:
                self.M = datetime.strptime(self.M, "%Y-%m-%dT%H:%M:%S")

        for variable_name in [
            "MO",
            "MC",
            "TO",
            "TC",
            "WO",
            "WC",
            "THO",
            "THC",
            "FO",
            "FC",
            "SO",
            "SC",
            "SUO",
            "SUC",
        ]:
            value = getattr(self, variable_name, None)
            if value is not None:
                if value.strip() == "":
                    setattr(self, variable_name, None)
                else:
                    setattr(
                        self,
                        variable_name,
                        datetime.strptime(value, "%H:%M").time(),
                    )

    def distance_between(self, latitude: float, longitude: float):
        """Function to return a distance between the site and a lat/long
        coordinate.
        :return: Distance object
        """
        distance = geopy.distance.geodesic(
            (self.Lat, self.Lng),
            (latitude, longitude),
        )
        return distance

    def is_open(self, date_time: datetime = None):
        if date_time is None:
            date_time = datetime.now()

        opening_hours = {
            "Monday": (self.MO, self.MC),
            "Tuesday": (self.TO, self.TC),
            "Wednesday": (self.WO, self.WC),
            "Thursday": (self.THO, self.THC),
            "Friday": (self.FO, self.FC),
            "Saturday": (self.SO, self.SC),
            "Sunday": (self.SUO, self.SUC),
        }

        weekday = date_time.strftime("%A")
        time = date_time.time()

        if (
            opening_hours[weekday][0] is not None
            and opening_hours[weekday][1] is not None
        ):
            if opening_hours[weekday][0] <= time <= opening_hours[weekday][1]:
                return True

        return False


@dataclass
class SitePrice:
    SiteId: int
    FuelId: int
    CollectionMethod: str
    TransactionDateUtc: str
    Price: float

    def __post_init__(self):
        if isinstance(self.TransactionDateUtc, str):
            self.TransactionDateUtc = parser.parse(self.TransactionDateUtc)
        if isinstance(self.Price, float):
            self.Price = Money(
                amount=Decimal(str(self.Price)),
                currency="AUD",
            )
